"""Google Cloud Platform compute instances state module.

Copyright (c) 2022 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0

"""

__contracts__ = ["resource"]

from dataclasses import field
from dataclasses import make_dataclass
from typing import Dict, Any, List
import copy
from idem_gcp.tool.gcp.utils import get_project_from_account


async def present(
    hub,
    ctx,
    name: str,
    zone: str,
    project: str = None,
    service_accounts: List[
        make_dataclass(
            "ServiceAccount",
            [
                ("scopes", List[str], field(default=None)),
                ("email", str, field(default=None)),
            ],
        )
    ] = None,
    label_fingerprint: str = None,
    fingerprint: str = None,
    advanced_machine_features: make_dataclass(
        "AdvancedMachineFeatures",
        [
            ("enable_nested_virtualization", bool, field(default=None)),
            ("visible_core_count", int, field(default=None)),
            ("enable_uefi_networking", bool, field(default=None)),
            ("threads_per_core", int, field(default=None)),
        ],
    ) = None,
    can_ip_forward: bool = None,
    network_interfaces: List[
        make_dataclass(
            "NetworkInterface",
            [
                ("kind", str, field(default=None)),
                ("network", str, field(default=None)),
                ("stack_type", str, field(default=None)),
                ("name", str, field(default=None)),
                ("subnetwork", str, field(default=None)),
                ("queue_count", int, field(default=None)),
                (
                    "ipv6_access_configs",
                    List[
                        make_dataclass(
                            "AccessConfig",
                            [
                                ("network_tier", str, field(default=None)),
                                (
                                    "external_ipv6_prefix_length",
                                    int,
                                    field(default=None),
                                ),
                                ("name", str, field(default=None)),
                                ("kind", str, field(default=None)),
                                ("external_ipv6", str, field(default=None)),
                                ("type_", str, field(default=None)),
                                ("set_public_ptr", bool, field(default=None)),
                                ("public_ptr_domain_name", str, field(default=None)),
                                ("nat_ip", str, field(default=None)),
                            ],
                        )
                    ],
                    field(default=None),
                ),
                ("internal_ipv6_prefix_length", int, field(default=None)),
                ("fingerprint", str, field(default=None)),
                (
                    "alias_ip_ranges",
                    List[
                        make_dataclass(
                            "AliasIpRange",
                            [
                                ("ip_cidr_range", str, field(default=None)),
                                ("subnetwork_range_name", str, field(default=None)),
                            ],
                        )
                    ],
                    field(default=None),
                ),
                ("ipv6_access_type", str, field(default=None)),
                ("network_ip", str, field(default=None)),
                ("nic_type", str, field(default=None)),
                (
                    "access_configs",
                    List[
                        make_dataclass(
                            "AccessConfig",
                            [
                                ("network_tier", str, field(default=None)),
                                (
                                    "external_ipv6_prefix_length",
                                    int,
                                    field(default=None),
                                ),
                                ("name", str, field(default=None)),
                                ("kind", str, field(default=None)),
                                ("external_ipv6", str, field(default=None)),
                                ("type_", str, field(default=None)),
                                ("set_public_ptr", bool, field(default=None)),
                                ("public_ptr_domain_name", str, field(default=None)),
                                ("nat_ip", str, field(default=None)),
                            ],
                        )
                    ],
                    field(default=None),
                ),
                ("ipv6_address", str, field(default=None)),
            ],
        )
    ] = None,
    labels: Dict[str, Any] = None,
    display_device: make_dataclass(
        "DisplayDevice", [("enable_display", bool, field(default=None))]
    ) = None,
    source_machine_image: str = None,
    resource_policies: List[str] = None,
    tags: make_dataclass(
        "Tags",
        [
            ("fingerprint", str, field(default=None)),
            ("items", List[str], field(default=None)),
        ],
    ) = None,
    key_revocation_action_type: str = None,
    scheduling: make_dataclass(
        "Scheduling",
        [
            ("on_host_maintenance", str, field(default=None)),
            ("preemptible", bool, field(default=None)),
            ("min_node_cpus", int, field(default=None)),
            ("automatic_restart", bool, field(default=None)),
            (
                "node_affinities",
                List[
                    make_dataclass(
                        "SchedulingNodeAffinity",
                        [
                            ("operator", str, field(default=None)),
                            ("values", List[str], field(default=None)),
                            ("key", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("instance_termination_action", str, field(default=None)),
            ("provisioning_model", str, field(default=None)),
            ("location_hint", str, field(default=None)),
        ],
    ) = None,
    private_ipv6_google_access: str = None,
    description: str = None,
    guest_accelerators: List[
        make_dataclass(
            "AcceleratorConfig",
            [
                ("accelerator_count", int, field(default=None)),
                ("accelerator_type", str, field(default=None)),
            ],
        )
    ] = None,
    confidential_instance_config: make_dataclass(
        "ConfidentialInstanceConfig",
        [("enable_confidential_compute", bool, field(default=None))],
    ) = None,
    shielded_instance_config: make_dataclass(
        "ShieldedInstanceConfig",
        [
            ("enable_vtpm", bool, field(default=None)),
            ("enable_secure_boot", bool, field(default=None)),
            ("enable_integrity_monitoring", bool, field(default=None)),
        ],
    ) = None,
    source_machine_image_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("sha256", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
            ("raw_key", str, field(default=None)),
        ],
    ) = None,
    disks: List[
        make_dataclass(
            "AttachedDisk",
            [
                ("disk_size_gb", str, field(default=None)),
                ("auto_delete", bool, field(default=None)),
                ("boot", bool, field(default=None)),
                (
                    "guest_os_features",
                    List[
                        make_dataclass(
                            "GuestOsFeature", [("type_", str, field(default=None))]
                        )
                    ],
                    field(default=None),
                ),
                ("source", str, field(default=None)),
                (
                    "disk_encryption_key",
                    make_dataclass(
                        "CustomerEncryptionKey",
                        [
                            ("kms_key_service_account", str, field(default=None)),
                            ("sha256", str, field(default=None)),
                            ("rsa_encrypted_key", str, field(default=None)),
                            ("kms_key_name", str, field(default=None)),
                            ("raw_key", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("force_attach", bool, field(default=None)),
                ("architecture", str, field(default=None)),
                ("device_name", str, field(default=None)),
                (
                    "shielded_instance_initial_state",
                    make_dataclass(
                        "InitialStateConfig",
                        [
                            (
                                "keks",
                                List[
                                    make_dataclass(
                                        "FileContentBuffer",
                                        [
                                            ("file_type", str, field(default=None)),
                                            ("content", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "dbs",
                                List[
                                    make_dataclass(
                                        "FileContentBuffer",
                                        [
                                            ("file_type", str, field(default=None)),
                                            ("content", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "dbxs",
                                List[
                                    make_dataclass(
                                        "FileContentBuffer",
                                        [
                                            ("file_type", str, field(default=None)),
                                            ("content", str, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "pk",
                                make_dataclass(
                                    "FileContentBuffer",
                                    [
                                        ("file_type", str, field(default=None)),
                                        ("content", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "initialize_params",
                    make_dataclass(
                        "AttachedDiskInitializeParams",
                        [
                            (
                                "resource_manager_tags",
                                Dict[str, Any],
                                field(default=None),
                            ),
                            ("source_image", str, field(default=None)),
                            ("description", str, field(default=None)),
                            ("disk_size_gb", str, field(default=None)),
                            ("source_snapshot", str, field(default=None)),
                            ("provisioned_iops", str, field(default=None)),
                            (
                                "source_image_encryption_key",
                                make_dataclass(
                                    "CustomerEncryptionKey",
                                    [
                                        (
                                            "kms_key_service_account",
                                            str,
                                            field(default=None),
                                        ),
                                        ("sha256", str, field(default=None)),
                                        ("rsa_encrypted_key", str, field(default=None)),
                                        ("kms_key_name", str, field(default=None)),
                                        ("raw_key", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("on_update_action", str, field(default=None)),
                            (
                                "source_snapshot_encryption_key",
                                make_dataclass(
                                    "CustomerEncryptionKey",
                                    [
                                        (
                                            "kms_key_service_account",
                                            str,
                                            field(default=None),
                                        ),
                                        ("sha256", str, field(default=None)),
                                        ("rsa_encrypted_key", str, field(default=None)),
                                        ("kms_key_name", str, field(default=None)),
                                        ("raw_key", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("architecture", str, field(default=None)),
                            ("licenses", List[str], field(default=None)),
                            ("labels", Dict[str, Any], field(default=None)),
                            ("disk_type", str, field(default=None)),
                            ("disk_name", str, field(default=None)),
                            ("resource_policies", List[str], field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("kind", str, field(default=None)),
                ("index", int, field(default=None)),
                ("interface", str, field(default=None)),
                ("licenses", List[str], field(default=None)),
                ("mode", str, field(default=None)),
                ("type_", str, field(default=None)),
            ],
        )
    ] = None,
    hostname: str = None,
    network_performance_config: make_dataclass(
        "NetworkPerformanceConfig",
        [("total_egress_bandwidth_tier", str, field(default=None))],
    ) = None,
    params: make_dataclass(
        "InstanceParams",
        [("resource_manager_tags", Dict[str, Any], field(default=None))],
    ) = None,
    deletion_protection: bool = None,
    metadata: make_dataclass(
        "Metadata",
        [
            (
                "items",
                List[
                    make_dataclass(
                        "Classcc3f9d3b673e4a89805fea2c3069bd85",
                        [
                            ("key", str, field(default=None)),
                            ("value", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("kind", str, field(default=None)),
            ("fingerprint", str, field(default=None)),
        ],
    ) = None,
    shielded_instance_integrity_policy: make_dataclass(
        "ShieldedInstanceIntegrityPolicy",
        [("update_auto_learn_policy", bool, field(default=None))],
    ) = None,
    reservation_affinity: make_dataclass(
        "ReservationAffinity",
        [
            ("key", str, field(default=None)),
            ("values", List[str], field(default=None)),
            ("consume_reservation_type", str, field(default=None)),
        ],
    ) = None,
    min_cpu_platform: str = None,
    machine_type: str = None,
    source_instance_template: str = None,
    request_id: str = None,
    minimal_action: str = None,
    most_disruptive_allowed_action: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Create or update a compute instance resource.

    Creates an instance resource in the specified project using the data included in the request.
    or
    Updates an instance only if the necessary resources are available. This method can update only a specific set of instance properties. See Updating a running instance for a list of updatable instance properties.
    Args:
        name(str):
            An Idem name of the resource.

        service_accounts(List[Dict[str, Any]], optional):
            A list of service accounts, with their specified scopes, authorized for this instance. Only one service account per VM instance is supported. Service accounts generate access tokens that can be accessed through the metadata server and used to authenticate applications on the instance. See Service Accounts for more information. Defaults to None.

                * scopes (List[str], optional):
                    The list of scopes to be made available for this service account.

                * email (str, optional):
                    Email address of the service account.

        label_fingerprint(str, optional):
            A fingerprint for this request, which is essentially a hash of the label's contents and used for optimistic locking. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update labels. You must always provide an up-to-date fingerprint hash in order to update or change labels. To see the latest fingerprint, make get() request to the instance. Defaults to None.

        fingerprint(str, optional):
            Specifies a fingerprint for this resource, which is essentially a hash of the instance's contents and used for optimistic locking. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update the instance. You must always provide an up-to-date fingerprint hash in order to update the instance. To see the latest fingerprint, make get() request to the instance. Defaults to None.

        advanced_machine_features(Dict[str, Any], optional):
            Controls for advanced machine-related behavior features.

                AdvancedMachineFeatures: Specifies options for controlling advanced machine features. Options that would traditionally be configured in a BIOS belong here. Features that require operating system support may have corresponding entries in the GuestOsFeatures of an Image (e.g., whether or not the OS in the Image supports nested virtualization being enabled or disabled). Defaults to None.
                * enable_nested_virtualization (bool, optional):
                    Whether to enable nested virtualization or not (default is false).

                * visible_core_count (int, optional):
                    The number of physical cores to expose to an instance. Multiply by the number of threads per core to compute the total number of virtual CPUs to expose to the instance. If unset, the number of cores is inferred from the instance's nominal CPU count and the underlying platform's SMT width.

                * enable_uefi_networking (bool, optional):
                    Whether to enable UEFI networking for instance creation.

                * threads_per_core (int, optional):
                    The number of threads per physical core. To disable simultaneous multithreading (SMT) set this to 1. If unset, the maximum number of threads supported per core by the underlying processor is assumed.
        can_ip_forward(bool, optional):
            Allows this instance to send and receive packets with non-matching destination or source IPs. This is required if you plan to use this instance to forward routes. For more information, see Enabling IP Forwarding . Defaults to None.

        network_interfaces(List[Dict[str, Any]], optional):
            An array of network configurations for this instance. These specify how interfaces are configured to interact with other network services, such as connecting to the internet. Multiple interfaces are supported per instance. Defaults to None.
                * kind (str, optional):
                    [Output Only] Type of the resource. Always compute#networkInterface for network interfaces.

                * network (str, optional):
                    URL of the VPC network resource for this instance. When creating an instance, if neither the network nor the subnetwork is specified, the default network global/networks/default is used. If the selected project doesn't have the default network, you must specify a network or subnet. If the network is not specified but the subnetwork is specified, the network is inferred. If you specify this property, you can specify the network as a full or partial URL. For example, the following are all valid URLs: - https://www.googleapis.com/compute/v1/projects/project/global/networks/ network - projects/project/global/networks/network - global/networks/default

                * stack_type (str, optional):
                    The stack type for this network interface to identify whether the IPv6 feature is enabled or not. If not specified, IPV4_ONLY will be used. This field can be both set at instance creation and update network interface operations.
                        Enum type. Allowed values:
                        "IPV4_IPV6" - The network interface can have both IPv4 and IPv6 addresses.
                        "IPV4_ONLY" - The network interface will be assigned IPv4 address.

                * name (str, optional):
                    [Output Only] The name of the network interface, which is generated by the server. For a VM, the network interface uses the nicN naming format. Where N is a value between 0 and 7. The default interface value is nic0.

                * subnetwork (str, optional):
                    The URL of the Subnetwork resource for this instance. If the network resource is in legacy mode, do not specify this field. If the network is in auto subnet mode, specifying the subnetwork is optional. If the network is in custom subnet mode, specifying the subnetwork is required. If you specify this field, you can specify the subnetwork as a full or partial URL. For example, the following are all valid URLs: - https://www.googleapis.com/compute/v1/projects/project/regions/region /subnetworks/subnetwork - regions/region/subnetworks/subnetwork

                * queue_count (int, optional):
                    The networking queue count that's specified by users for the network interface. Both Rx and Tx queues will be set to this number. It'll be empty if not specified by the users.

                * ipv6_access_configs (List[Dict[str, Any]], optional):
                    An array of IPv6 access configurations for this interface. Currently, only one IPv6 access config, DIRECT_IPV6, is supported. If there is no ipv6AccessConfig specified, then this instance will have no external IPv6 Internet access.

                        * network_tier (str, optional):
                            This signifies the networking tier used for configuring this access configuration and can only take the following values: PREMIUM, STANDARD. If an AccessConfig is specified without a valid external IP address, an ephemeral IP will be created with this networkTier. If an AccessConfig with a valid external IP address is specified, it must match that of the networkTier associated with the Address resource owning that IP.
                                Enum type. Allowed values:
                                "FIXED_STANDARD" - Public internet quality with fixed bandwidth.
                                "PREMIUM" - High quality, Google-grade network tier, support for all networking products.
                                "STANDARD" - Public internet quality, only limited support for other networking products.
                                "STANDARD_OVERRIDES_FIXED_STANDARD" - (Output only) Temporary tier for FIXED_STANDARD when fixed standard tier is expired or not configured.

                        * external_ipv6_prefix_length (int, optional):
                            The prefix length of the external IPv6 range.

                        * name (str, optional):
                            The name of this access configuration. The default and recommended name is External NAT, but you can use any arbitrary string, such as My external IP or Network Access.

                        * kind (str, optional):
                            [Output Only] Type of the resource. Always compute#accessConfig for access configs.

                        * external_ipv6 (str, optional):
                            The first IPv6 address of the external IPv6 range associated with this instance, prefix length is stored in externalIpv6PrefixLength in ipv6AccessConfig. The field is output only, an IPv6 address from a subnetwork associated with the instance will be allocated dynamically.

                        * type_ (str, optional):
                            The type of configuration. The default and only option is ONE_TO_ONE_NAT.
                                Enum type. Allowed values:
                                "DIRECT_IPV6"
                                "ONE_TO_ONE_NAT"

                        * set_public_ptr (bool, optional):
                            Specifies whether a public DNS 'PTR' record should be created to map the external IP address of the instance to a DNS domain name. This field is not used in ipv6AccessConfig. A default PTR record will be created if the VM has external IPv6 range associated.

                        * public_ptr_domain_name (str, optional):
                            The DNS domain name for the public PTR record. You can set this field only if the `setPublicPtr` field is enabled in accessConfig. If this field is unspecified in ipv6AccessConfig, a default PTR record will be createc for first IP in associated external IPv6 range.

                        * nat_ip (str, optional):
                            An external IP address associated with this instance. Specify an unused static external IP address available to the project or leave this field undefined to use an IP from a shared ephemeral IP address pool. If you specify a static external IP address, it must live in the same region as the zone of the instance.

                * internal_ipv6_prefix_length (int, optional):
                    The prefix length of the primary internal IPv6 range.

                * fingerprint (str, optional):
                    Fingerprint hash of contents stored in this network interface. This field will be ignored when inserting an Instance or adding a NetworkInterface. An up-to-date fingerprint must be provided in order to update the NetworkInterface. The request will fail with error 400 Bad Request if the fingerprint is not provided, or 412 Precondition Failed if the fingerprint is out of date.

                * alias_ip_ranges (List[Dict[str, Any]], optional):
                    An array of alias IP ranges for this network interface. You can only specify this field for network interfaces in VPC networks.

                        * ip_cidr_range (str, optional):
                            The IP alias ranges to allocate for this interface. This IP CIDR range must belong to the specified subnetwork and cannot contain IP addresses reserved by system or used by other network interfaces. This range may be a single IP address (such as 10.2.3.4), a netmask (such as /24) or a CIDR-formatted string (such as 10.1.2.0/24).

                        * subnetwork_range_name (str, optional):
                            The name of a subnetwork secondary IP range from which to allocate an IP alias range. If not specified, the primary range of the subnetwork is used.

                * ipv6_access_type (str, optional):
                    [Output Only] One of EXTERNAL, INTERNAL to indicate whether the IP can be accessed from the Internet. This field is always inherited from its subnetwork. Valid only if stackType is IPV4_IPV6.
                        Enum type. Allowed values:
                        "EXTERNAL" - This network interface can have external IPv6.
                        "INTERNAL" - This network interface can have internal IPv6.

                * network_ip (str, optional):
                    An IPv4 internal IP address to assign to the instance for this network interface. If not specified by the user, an unused internal IP is assigned by the system.

                * nic_type (str, optional):
                    The type of vNIC to be used on this interface. This may be gVNIC or VirtioNet.
                        Enum type. Allowed values:
                        "GVNIC" - GVNIC
                        "UNSPECIFIED_NIC_TYPE" - No type specified.
                        "VIRTIO_NET" - VIRTIO

                * access_configs (List[Dict[str, Any]], optional):
                    An array of configurations for this interface. Currently, only one access config, ONE_TO_ONE_NAT, is supported. If there are no accessConfigs specified, then this instance will have no external internet access.

                        * network_tier (str, optional):
                            This signifies the networking tier used for configuring this access configuration and can only take the following values: PREMIUM, STANDARD. If an AccessConfig is specified without a valid external IP address, an ephemeral IP will be created with this networkTier. If an AccessConfig with a valid external IP address is specified, it must match that of the networkTier associated with the Address resource owning that IP.
                                Enum type. Allowed values:
                                "FIXED_STANDARD" - Public internet quality with fixed bandwidth.
                                "PREMIUM" - High quality, Google-grade network tier, support for all networking products.
                                "STANDARD" - Public internet quality, only limited support for other networking products.
                                "STANDARD_OVERRIDES_FIXED_STANDARD" - (Output only) Temporary tier for FIXED_STANDARD when fixed standard tier is expired or not configured.

                        * external_ipv6_prefix_length (int, optional):
                            The prefix length of the external IPv6 range.

                        * name (str, optional):
                            The name of this access configuration. The default and recommended name is External NAT, but you can use any arbitrary string, such as My external IP or Network Access.

                        * kind (str, optional):
                            [Output Only] Type of the resource. Always compute#accessConfig for access configs.

                        * external_ipv6 (str, optional):
                            The first IPv6 address of the external IPv6 range associated with this instance, prefix length is stored in externalIpv6PrefixLength in ipv6AccessConfig. The field is output only, an IPv6 address from a subnetwork associated with the instance will be allocated dynamically.

                        * type_ (str, optional):
                            The type of configuration. The default and only option is ONE_TO_ONE_NAT.
                                Enum type. Allowed values:
                                "DIRECT_IPV6"
                                "ONE_TO_ONE_NAT"

                        * set_public_ptr (bool, optional):
                            Specifies whether a public DNS 'PTR' record should be created to map the external IP address of the instance to a DNS domain name. This field is not used in ipv6AccessConfig. A default PTR record will be created if the VM has external IPv6 range associated.

                        * public_ptr_domain_name (str, optional):
                            The DNS domain name for the public PTR record. You can set this field only if the `setPublicPtr` field is enabled in accessConfig. If this field is unspecified in ipv6AccessConfig, a default PTR record will be createc for first IP in associated external IPv6 range.

                        * nat_ip (str, optional):
                            An external IP address associated with this instance. Specify an unused static external IP address available to the project or leave this field undefined to use an IP from a shared ephemeral IP address pool. If you specify a static external IP address, it must live in the same region as the zone of the instance.

                * ipv6_address (str, optional):
                    An IPv6 internal network address for this network interface.

        labels(Dict[str, Any], optional):
            Labels to apply to this instance. These can be later modified by the setLabels method. Defaults to None.

        display_device(Dict[str, Any], optional):
            Enables display device for the instance.
                DisplayDevice: A set of Display Device options. Defaults to None.

                * enable_display (bool, optional):
                    Defines whether the instance has Display enabled.

        source_machine_image(str, optional):
            Source machine image. Defaults to None.

        resource_policies(List[str], optional):
            Resource policies applied to this instance. Defaults to None.

        tags(Dict[str, Any], optional):
            Tags to apply to this instance. Tags are used to identify valid sources or targets for network firewalls and are specified by the client during instance creation. The tags can be later modified by the setTags method. Each tag within the list must comply with RFC1035. Multiple tags can be specified via the 'tags.items' field.
                Tags: A set of instance tags. Defaults to None.

                * fingerprint (str, optional):
                    Specifies a fingerprint for this request, which is essentially a hash of the tags' contents and used for optimistic locking. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update tags. You must always provide an up-to-date fingerprint hash in order to update or change tags. To see the latest fingerprint, make get() request to the instance.

                * items (List[str], optional):
                    An array of tags. Each tag must be 1-63 characters long, and comply with RFC1035.

        key_revocation_action_type(str, optional):
            KeyRevocationActionType of the instance. Supported options are "STOP" and "NONE". The default value is "NONE" if it is not specified.
                Enum type. Allowed values:
                "KEY_REVOCATION_ACTION_TYPE_UNSPECIFIED" - Default value. This value is unused.
                "NONE" - Indicates user chose no operation.
                "STOP" - Indicates user chose to opt for VM shutdown on key revocation. Defaults to None.

        scheduling(Dict[str, Any], optional):
            Sets the scheduling options for this instance.
                Scheduling: Sets the scheduling options for an Instance. Defaults to None.

                * on_host_maintenance (str, optional):
                    Defines the maintenance behavior for this instance. For standard instances, the default behavior is MIGRATE. For preemptible instances, the default and only possible behavior is TERMINATE. For more information, see Set VM host maintenance policy.
                        Enum type. Allowed values:
                        "MIGRATE" - *[Default]* Allows Compute Engine to automatically migrate instances out of the way of maintenance events.
                        "TERMINATE" - Tells Compute Engine to terminate and (optionally) restart the instance away from the maintenance activity. If you would like your instance to be restarted, set the automaticRestart flag to true. Your instance may be restarted more than once, and it may be restarted outside the window of maintenance events.

                * preemptible (bool, optional):
                    Defines whether the instance is preemptible. This can only be set during instance creation or while the instance is stopped and therefore, in a `TERMINATED` state. See Instance Life Cycle for more information on the possible instance states.

                * min_node_cpus (int, optional):
                    The minimum number of virtual CPUs this instance will consume when running on a sole-tenant node.

                * automatic_restart (bool, optional):
                    Specifies whether the instance should be automatically restarted if it is terminated by Compute Engine (not terminated by a user). You can only set the automatic restart option for standard instances. Preemptible instances cannot be automatically restarted. By default, this is set to true so an instance is automatically restarted if it is terminated by Compute Engine.

                * node_affinities (List[Dict[str, Any]], optional):
                    A set of node affinity and anti-affinity configurations. Refer to Configuring node affinity for more information. Overrides reservationAffinity.

                    * operator (str, optional):
                        Defines the operation of node selection. Valid operators are IN for affinity and NOT_IN for anti-affinity.
                            Enum type. Allowed values:
                            "IN" - Requires Compute Engine to seek for matched nodes.
                            "NOT_IN" - Requires Compute Engine to avoid certain nodes.
                            "OPERATOR_UNSPECIFIED"

                    * values (List[str], optional):
                        Corresponds to the label values of Node resource.

                    * key (str, optional):
                        Corresponds to the label key of Node resource.

                * instance_termination_action (str, optional):
                    Specifies the termination action for the instance.
                        Enum type. Allowed values:
                        "DELETE" - Delete the VM.
                        "INSTANCE_TERMINATION_ACTION_UNSPECIFIED" - Default value. This value is unused.
                        "STOP" - Stop the VM without storing in-memory content. default action.

                * provisioning_model (str, optional):
                    Specifies the provisioning model of the instance.
                        Enum type. Allowed values:
                        "SPOT" - Heavily discounted, no guaranteed runtime.
                        "STANDARD" - Standard provisioning with user controlled runtime, no discounts.

                * location_hint (str, optional):
                    An opaque location hint used to place the instance close to other resources. This field is for use by internal tools that use the public API.

        private_ipv6_google_access(str, optional):
            The private IPv6 google access type for the VM. If not specified, use INHERIT_FROM_SUBNETWORK as default.
                Enum type. Allowed values:
                "ENABLE_BIDIRECTIONAL_ACCESS_TO_GOOGLE" - Bidirectional private IPv6 access to/from Google services. If specified, the subnetwork who is attached to the instance's default network interface will be assigned an internal IPv6 prefix if it doesn't have before.
                "ENABLE_OUTBOUND_VM_ACCESS_TO_GOOGLE" - Outbound private IPv6 access from VMs in this subnet to Google services. If specified, the subnetwork who is attached to the instance's default network interface will be assigned an internal IPv6 prefix if it doesn't have before.
                "INHERIT_FROM_SUBNETWORK" - Each network interface inherits PrivateIpv6GoogleAccess from its subnetwork. Defaults to None.

        description(str, optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        guest_accelerators(List[Dict[str, Any]], optional):
            A list of the type and count of accelerator cards attached to the instance. Defaults to None.

                * accelerator_count (int, optional):
                    The number of the guest accelerator cards exposed to this instance.

                * accelerator_type (str, optional):
                    Full or partial URL of the accelerator type resource to attach to this instance. For example: projects/my-project/zones/us-central1-c/acceleratorTypes/nvidia-tesla-p100 If you are creating an instance template, specify only the accelerator name. See GPUs on Compute Engine for a full list of accelerator types.

        confidential_instance_config(Dict[str, Any], optional):
            ConfidentialInstanceConfig: A set of Confidential Instance options. Defaults to None.

            * enable_confidential_compute (bool, optional):
                Defines whether the instance should have confidential compute enabled.

        shielded_instance_config(Dict[str, Any], optional):
            ShieldedInstanceConfig: A set of Shielded Instance options. Defaults to None.

                * enable_vtpm (bool, optional):
                    Defines whether the instance has the vTPM enabled. Enabled by default.

                * enable_secure_boot (bool, optional):
                    Defines whether the instance has Secure Boot enabled. Disabled by default.

                * enable_integrity_monitoring (bool, optional):
                    Defines whether the instance has integrity monitoring enabled. Enabled by default.

        source_machine_image_encryption_key(Dict[str, Any], optional):
            Source machine image encryption key when creating an instance from a machine image. Defaults to None.

                * kms_key_service_account (str, optional):
                    The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/

                * sha256 (str, optional):
                    [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.

                * rsa_encrypted_key (str, optional):
                    Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem

                * kms_key_name (str, optional):
                    The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key

                * raw_key (str, optional):
                    Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

        disks(List[Dict[str, Any]], optional):
            Array of disks associated with this instance. Persistent disks must be created before you can assign them. Defaults to None.

                * disk_size_gb (str, optional):
                    The size of the disk in GB.

                * auto_delete (bool, optional):
                    Specifies whether the disk will be auto-deleted when the instance is deleted (but not when the disk is detached from the instance).

                * boot (bool, optional):
                    Indicates that this is a boot disk. The virtual machine will use the first partition of the disk for its root filesystem.

                * guest_os_features (List[Dict[str, Any]], optional):
                    A list of features to enable on the guest operating system. Applicable only for bootable images. Read Enabling guest operating system features to see a list of available options.

                        * type_ (str, optional):
                            The ID of a supported feature. To add multiple values, use commas to separate values. Set to one or more of the following values: - VIRTIO_SCSI_MULTIQUEUE - WINDOWS - MULTI_IP_SUBNET - UEFI_COMPATIBLE - GVNIC - SEV_CAPABLE - SUSPEND_RESUME_COMPATIBLE - SEV_SNP_CAPABLE For more information, see Enabling guest operating system features.
                                Enum type. Allowed values:
                                "FEATURE_TYPE_UNSPECIFIED"
                                "GVNIC"
                                "MULTI_IP_SUBNET"
                                "SECURE_BOOT"
                                "SEV_CAPABLE"
                                "UEFI_COMPATIBLE"
                                "VIRTIO_SCSI_MULTIQUEUE"
                                "WINDOWS"

                * source (str, optional):
                    Specifies a valid partial or full URL to an existing Persistent Disk resource. When creating a new instance, one of initializeParams.sourceImage or initializeParams.sourceSnapshot or disks.source is required except for local SSD. If desired, you can also attach existing non-root persistent disks using this property. This field is only applicable for persistent disks. Note that for InstanceTemplate, specify the disk name for zonal disk, and the URL for regional disk.

                * disk_encryption_key (Dict[str, Any], optional):
                    Encrypts or decrypts a disk using a customer-supplied encryption key. If you are creating a new disk, this field encrypts the new disk using an encryption key that you provide. If you are attaching an existing disk that is already encrypted, this field decrypts the disk using the customer-supplied encryption key. If you encrypt a disk using a customer-supplied key, you must provide the same key again when you attempt to use this resource at a later time. For example, you must provide the key when you create a snapshot or an image from the disk or when you attach the disk to a virtual machine instance. If you do not provide an encryption key, then the disk will be encrypted using an automatically generated key and you do not need to provide a key to use the disk later. Instance templates do not store customer-supplied encryption keys, so you cannot use your own keys to encrypt disks in a managed instance group.

                        * kms_key_service_account (str, optional):
                            The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/

                        * sha256 (str, optional):
                            [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.

                        * rsa_encrypted_key (str, optional):
                            Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem

                        * kms_key_name (str, optional):
                            The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key

                        * raw_key (str, optional):
                            Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

                * force_attach (bool, optional):
                    [Input Only] Whether to force attach the regional disk even if it's currently attached to another instance. If you try to force attach a zonal disk to an instance, you will receive an error.

                * architecture (str, optional):
                    [Output Only] The architecture of the attached disk. Valid values are ARM64 or X86_64.
                        Enum type. Allowed values:
                        "ARCHITECTURE_UNSPECIFIED" - Default value indicating Architecture is not set.
                        "ARM64" - Machines with architecture ARM64
                        "X86_64" - Machines with architecture X86_64

                * device_name (str, optional):
                    Specifies a unique device name of your choice that is reflected into the /dev/disk/by-id/google-* tree of a Linux operating system running within the instance. This name can be used to reference the device for mounting, resizing, and so on, from within the instance. If not specified, the server chooses a default device name to apply to this disk, in the form persistent-disk-x, where x is a number assigned by Google Compute Engine. This field is only applicable for persistent disks.

                * shielded_instance_initial_state (Dict[str, Any], optional):
                    [Output Only] shielded vm initial state stored on disk
                        InitialStateConfig: Initial State for shielded instance, these are public keys which are safe to store in public

                        * keks (List[Dict[str, Any]], optional):
                            The Key Exchange Key (KEK).

                                * file_type (str, optional):
                                    The file type of source file.
                                        Enum type. Allowed values:
                                        "BIN"
                                        "UNDEFINED"
                                        "X509"

                                * content (str, optional):
                                    The raw content in the secure keys file.

                        * dbs (List[Dict[str, Any]], optional):
                            The Key Database (db).

                                * file_type (str, optional):
                                    The file type of source file.
                                        Enum type. Allowed values:
                                        "BIN"
                                        "UNDEFINED"
                                        "X509"

                                * content (str, optional):
                                    The raw content in the secure keys file.

                        * dbxs (List[Dict[str, Any]], optional):
                            The forbidden key database (dbx).

                                * file_type (str, optional):
                                    The file type of source file.
                                        Enum type. Allowed values:
                                        "BIN"
                                        "UNDEFINED"
                                        "X509"

                                * content (str, optional):
                                    The raw content in the secure keys file.

                        * pk (Dict[str, Any], optional):
                            The Platform Key (PK).

                                * file_type (str, optional):
                                    The file type of source file.
                                        Enum type. Allowed values:
                                        "BIN"
                                        "UNDEFINED"
                                        "X509"

                                * content (str, optional):
                                    The raw content in the secure keys file.

                * initialize_params (Dict[str, Any], optional):
                    [Input Only] Specifies the parameters for a new disk that will be created alongside the new instance. Use initialization parameters to create boot disks or local SSDs attached to the new instance. This property is mutually exclusive with the source property; you can only define one or the other, but not both.
                        AttachedDiskInitializeParams: [Input Only] Specifies the parameters for a new disk that will be created alongside the new instance. Use initialization parameters to create boot disks or local SSDs attached to the new instance. This field is persisted and returned for instanceTemplate and not returned in the context of instance. This property is mutually exclusive with the source property; you can only define one or the other, but not both.

                        * resource_manager_tags (Dict[str, Any], optional):
                            Resource manager tags to be bound to the disk. Tag keys and values have the same definition as resource manager tags. Keys must be in the format `tagKeys/{tag_key_id}`, and values are in the format `tagValues/456`. The field is ignored (both PUT & PATCH) when empty.

                        * source_image (str, optional):
                            The source image to create this disk. When creating a new instance, one of initializeParams.sourceImage or initializeParams.sourceSnapshot or disks.source is required except for local SSD. To create a disk with one of the public operating system images, specify the image by its family name. For example, specify family/debian-9 to use the latest Debian 9 image: projects/debian-cloud/global/images/family/debian-9 Alternatively, use a specific version of a public operating system image: projects/debian-cloud/global/images/debian-9-stretch-vYYYYMMDD To create a disk with a custom image that you created, specify the image name in the following format: global/images/my-custom-image You can also specify a custom image by its image family, which returns the latest version of the image in that family. Replace the image name with family/family-name: global/images/family/my-image-family If the source image is deleted later, this field will not be set.

                        * description (str, optional):
                            An optional description. Provide this property when creating the disk.

                        * disk_size_gb (str, optional):
                            Specifies the size of the disk in base-2 GB. The size must be at least 10 GB. If you specify a sourceImage, which is required for boot disks, the default size is the size of the sourceImage. If you do not specify a sourceImage, the default disk size is 500 GB.

                        * source_snapshot (str, optional):
                            The source snapshot to create this disk. When creating a new instance, one of initializeParams.sourceSnapshot or initializeParams.sourceImage or disks.source is required except for local SSD. To create a disk with a snapshot that you created, specify the snapshot name in the following format: global/snapshots/my-backup If the source snapshot is deleted later, this field will not be set.

                        * provisioned_iops (str, optional):
                            Indicates how many IOPS to provision for the disk. This sets the number of I/O operations per second that the disk can handle. Values must be between 10,000 and 120,000. For more details, see the Extreme persistent disk documentation.

                        * source_image_encryption_key (Dict[str, Any], optional):
                            The customer-supplied encryption key of the source image. Required if the source image is protected by a customer-supplied encryption key. InstanceTemplate and InstancePropertiesPatch do not store customer-supplied encryption keys, so you cannot create disks for instances in a managed instance group if the source images are encrypted with your own keys.

                                * kms_key_service_account (str, optional):
                                    The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/

                                * sha256 (str, optional):
                                    [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.

                                * rsa_encrypted_key (str, optional):
                                    Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem

                                * kms_key_name (str, optional):
                                    The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key

                                * raw_key (str, optional):
                                    Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

                        * on_update_action (str, optional):
                            Specifies which action to take on instance update with this disk. Default is to use the existing disk.
                                Enum type. Allowed values:
                                "RECREATE_DISK" - Always recreate the disk.
                                "RECREATE_DISK_IF_SOURCE_CHANGED" - Recreate the disk if source (image, snapshot) of this disk is different from source of existing disk.
                                "USE_EXISTING_DISK" - Use the existing disk, this is the default behaviour.

                        * source_snapshot_encryption_key (Dict[str, Any], optional):
                            The customer-supplied encryption key of the source snapshot.

                                * kms_key_service_account (str, optional):
                                    The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/

                                * sha256 (str, optional):
                                    [Output only] The RFC 4648 base64 encoded SHA-256 hash of the customer-supplied encryption key that protects this resource.

                                * rsa_encrypted_key (str, optional):
                                    Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem

                                * kms_key_name (str, optional):
                                    The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key

                                * raw_key (str, optional):
                                    Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

                        * architecture (str, optional):
                            The architecture of the attached disk. Valid values are arm64 or x86_64.
                                Enum type. Allowed values:
                                "ARCHITECTURE_UNSPECIFIED" - Default value indicating Architecture is not set.
                                "ARM64" - Machines with architecture ARM64
                                "X86_64" - Machines with architecture X86_64

                        * licenses (List[str], optional):
                            A list of publicly visible licenses. Reserved for Google's use.

                        * labels (Dict[str, Any], optional):
                            Labels to apply to this disk. These can be later modified by the disks.setLabels method. This field is only applicable for persistent disks.

                        * disk_type (str, optional):
                            Specifies the disk type to use to create the instance. If not specified, the default is pd-standard, specified using the full URL. For example: https://www.googleapis.com/compute/v1/projects/project/zones/zone /diskTypes/pd-standard For a full list of acceptable values, see Persistent disk types. If you specify this field when creating a VM, you can provide either the full or partial URL. For example, the following values are valid: - https://www.googleapis.com/compute/v1/projects/project/zones/zone /diskTypes/diskType - projects/project/zones/zone/diskTypes/diskType - zones/zone/diskTypes/diskType If you specify this field when creating or updating an instance template or all-instances configuration, specify the type of the disk, not the URL. For example: pd-standard.

                        * disk_name (str, optional):
                            Specifies the disk name. If not specified, the default is to use the name of the instance. If a disk with the same name already exists in the given region, the existing disk is attached to the new instance and the new disk is not created.

                        * resource_policies (List[str], optional):
                            Resource policies applied to this disk for automatic snapshot creations. Specified using the full or partial URL. For instance template, specify only the resource policy name.

                * kind (str, optional):
                    [Output Only] Type of the resource. Always compute#attachedDisk for attached disks.

                * index (int, optional):
                    [Output Only] A zero-based index to this disk, where 0 is reserved for the boot disk. If you have many disks attached to an instance, each disk would have a unique index number.

                * interface (str, optional):
                    Specifies the disk interface to use for attaching this disk, which is either SCSI or NVME. For most machine types, the default is SCSI. Local SSDs can use either NVME or SCSI. In certain configurations, persistent disks can use NVMe. For more information, see About persistent disks.
                        Enum type. Allowed values:
                        "NVME"
                        "SCSI"

                * licenses (List[str], optional):
                    [Output Only] Any valid publicly visible licenses.

                * mode (str, optional):
                    The mode in which to attach this disk, either READ_WRITE or READ_ONLY. If not specified, the default is to attach the disk in READ_WRITE mode.
                        Enum type. Allowed values:
                        "READ_ONLY" - Attaches this disk in read-only mode. Multiple virtual machines can use a disk in read-only mode at a time.
                        "READ_WRITE" - *[Default]* Attaches this disk in read-write mode. Only one virtual machine at a time can be attached to a disk in read-write mode.

                * type_ (str, optional):
                    Specifies the type of the disk, either SCRATCH or PERSISTENT. If not specified, the default is PERSISTENT.
                        Enum type. Allowed values:
                        "PERSISTENT"
                        "SCRATCH"

        hostname(str, optional):
            Specifies the hostname of the instance. The specified hostname must be RFC1035 compliant. If hostname is not specified, the default hostname is [INSTANCE_NAME].c.[PROJECT_ID].internal when using the global DNS, and [INSTANCE_NAME].[ZONE].c.[PROJECT_ID].internal when using zonal DNS. Defaults to None.

        network_performance_config(Dict[str, Any], optional):
            . Defaults to None.

                * total_egress_bandwidth_tier (str, optional):
                    Enum type. Allowed values:
                        "DEFAULT"
                        "TIER_1"

        params(Dict[str, Any], optional):
            Input only. [Input Only] Additional params passed with the request, but not persisted as part of resource payload.
                InstanceParams: Additional instance params. Defaults to None.

                * resource_manager_tags (Dict[str, Any], optional):
                    Resource manager tags to be bound to the instance. Tag keys and values have the same definition as resource manager tags. Keys must be in the format `tagKeys/{tag_key_id}`, and values are in the format `tagValues/456`. The field is ignored (both PUT & PATCH) when empty.

                ResourceStatus: Contains output only fields. Use this sub-message for actual values set on Instance attributes as compared to the value requested by the user (intent) in their instance CRUD calls. Defaults to None.
                * physical_host (str, optional):
                    [Output Only] An opaque ID of the host on which the VM is running.

        deletion_protection(bool, optional):
            Whether the resource should be protected against deletion. Defaults to None.

        metadata(Dict[str, Any], optional):
            The metadata key/value pairs assigned to this instance. This includes custom metadata and predefined keys.
            Metadata: A metadata key/value entry. Defaults to None.

                * items (List[Dict[str, Any]], optional):
                    Array of key/value pairs. The total size of all keys and values must be less than 512 KB.

                        * key (str, optional):
                            Key for the metadata entry. Keys must conform to the following regexp: [a-zA-Z0-9-_]+, and be less than 128 bytes in length. This is reflected as part of a URL in the metadata server. Additionally, to avoid ambiguity, keys must not conflict with any other metadata keys for the project.

                        * value (str, optional):
                            Value for the metadata entry. These are free-form strings, and only have meaning as interpreted by the image running in the instance. The only restriction placed on values is that their size must be less than or equal to 262144 bytes (256 KiB).

                * kind (str, optional):
                    [Output Only] Type of the resource. Always compute#metadata for metadata.

                * fingerprint (str, optional):
                    Specifies a fingerprint for this request, which is essentially a hash of the metadata's contents and used for optimistic locking. The fingerprint is initially generated by Compute Engine and changes after every request to modify or update metadata. You must always provide an up-to-date fingerprint hash in order to update or change metadata, otherwise the request will fail with error 412 conditionNotMet. To see the latest fingerprint, make a get() request to retrieve the resource.

        zone(str):
            The name of the zone for this request.

        shielded_instance_integrity_policy(Dict[str, Any], optional):
            ShieldedInstanceIntegrityPolicy: The policy describes the baseline against which Instance boot integrity is measured. Defaults to None.

                * update_auto_learn_policy (bool, optional):
                    Updates the integrity policy baseline using the measurements from the VM instance's most recent boot.

        reservation_affinity(Dict[str, Any], optional):
            Specifies the reservations that this instance can consume from.
            ReservationAffinity: Specifies the reservations that this instance can consume from. Defaults to None.

                * key (str, optional):
                    Corresponds to the label key of a reservation resource. To target a SPECIFIC_RESERVATION by name, specify googleapis.com/reservation-name as the key and specify the name of your reservation as its value.

                * values (List[str], optional):
                    Corresponds to the label values of a reservation resource. This can be either a name to a reservation in the same project or "projects/different-project/reservations/some-reservation-name" to target a shared reservation in the same zone but in a different project.

                * consume_reservation_type (str, optional):
                    Specifies the type of reservation from which this instance can consume resources: ANY_RESERVATION (default), SPECIFIC_RESERVATION, or NO_RESERVATION. See Consuming reserved instances for examples.
                        Enum type. Allowed values:
                        "ANY_RESERVATION" - Consume any allocation available.
                        "NO_RESERVATION" - Do not consume from any allocated capacity.
                        "SPECIFIC_RESERVATION" - Must consume from a specific reservation. Must specify key value fields for specifying the reservations.
                        "UNSPECIFIED"

        min_cpu_platform(str, optional):
            Specifies a minimum CPU platform for the VM instance. Applicable values are the friendly names of CPU platforms, such as minCpuPlatform: "Intel Haswell" or minCpuPlatform: "Intel Sandy Bridge". Defaults to None.

        machine_type(str, optional):
            Full or partial URL of the machine type resource to use for this instance, in the format: zones/zone/machineTypes/machine-type. This is provided by the client when the instance is created. For example, the following is a valid partial url to a predefined machine type: zones/us-central1-f/machineTypes/n1-standard-1 To create a custom machine type, provide a URL to a machine type in the following format, where CPUS is 1 or an even number up to 32 (2, 4, 6, ... 24, etc), and MEMORY is the total memory for this instance. Memory must be a multiple of 256 MB and must be supplied in MB (e.g. 5 GB of memory is 5120 MB): zones/zone/machineTypes/custom-CPUS-MEMORY For example: zones/us-central1-f/machineTypes/custom-4-5120 For a full list of restrictions, read the Specifications for custom machine types. Defaults to None.

        project(str, Optional):
            Project ID for this request.

        source_instance_template(str, optional):
            Specifies instance template to create the instance. This field is optional. It can be a full or partial URL. For example, the following are all valid URLs to an instance template: - https://www.googleapis.com/compute/v1/projects/project /global/instanceTemplates/instanceTemplate - projects/project/global/instanceTemplates/instanceTemplate - global/instanceTemplates/instanceTemplate . Defaults to None.

        source_machine_image(str, optional):
            Specifies the machine image to use to create the instance. This field is optional. It can be a full or partial URL. For example, the following are all valid URLs to a machine image: - https://www.googleapis.com/compute/v1/projects/project/global/global /machineImages/machineImage - projects/project/global/global/machineImages/machineImage - global/machineImages/machineImage . Defaults to None.

        request_id(str, optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        minimal_action(str, optional):
            Specifies the action to take when updating an instance even if the updated properties do not require it. If not specified, then Compute Engine acts based on the minimum action that the updated properties require. Defaults to None.

        most_disruptive_allowed_action(str, optional):
            Specifies the most disruptive action that can be taken on the instance as part of the update. Compute Engine returns an error if the instance properties require a more disruptive action as part of the instance update. Valid options from lowest to highest are NO_EFFECT, REFRESH, and RESTART. Defaults to None.

        resource_id(str, optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        example_resource_name:
          gcp.compute.instance.present:
          - project: example-project
          - machine_type: https://www.googleapis.com/compute/v1/projects/example-project/zones/us-central1-a/machineTypes/g1-small
          - zone: us-central1-a
          - can_ip_forward: false
          - network_interfaces:
            - access_configs:
              - kind: compute#accessConfig
                name: External NAT
                network_tier: PREMIUM
                set_public_ptr: false
                type: ONE_TO_ONE_NAT
              kind: compute#networkInterface
              name: nic0
              network: https://www.googleapis.com/compute/v1/projects/project-name/global/networks/default
              stack_type: IPV4_ONLY
              subnetwork: https://www.googleapis.com/compute/v1/projects/project-name/regions/us-central1/subnetworks/default
          - disks:
            - auto_delete: true
              boot: true
              device_name: example_disk_1
              source: https://www.googleapis.com/compute/v1/projects/project-name/zones/us-central1-a/disks/example_disk_1
              mode: READ_WRITE
              type: PERSISTENT
              disk_size_gb: '10'
              index: 0
              interface: SCSI
              kind: compute#attachedDisk
          - scheduling:
              automatic_restart: true
              on_host_maintenance: MIGRATE
              preemptible: false
              provisioning_model: STANDARD
          - deletion_protection: false
          - tags:
              items:
                - test
          - metadata:
              kind: compute#metadata
              items:
                - key: sample_metadata_key
                  value: sample_metadata_value

    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }
    project = get_project_from_account(ctx, project)

    if ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.instance"
        )

        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    if resource_id:
        old_get_ret = await hub.exec.gcp.compute.instance.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"] or not old_get_ret["ret"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        result["old_state"] = copy.deepcopy(copy.copy(old_get_ret["ret"]))
    elif not hub.OPT.idem.get("get_resource_only_with_resource_id", False):
        resource_id = (ctx.get("old_state") or {}).get(
            "resource_id"
        ) or hub.tool.gcp.resource_prop_utils.construct_resource_id(
            "compute.instance", {**locals(), "instance": name}
        )
        old_get_ret = await hub.exec.gcp.compute.instance.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        if old_get_ret["ret"]:
            result["old_state"] = copy.deepcopy(copy.copy(old_get_ret["ret"]))

    # to be autogenerated by pop-create based on insert/update props in properties.yaml
    resource_body = {
        "service_accounts": service_accounts,
        "name": name,
        "min_cpu_platform": min_cpu_platform,
        "disks": disks,
        "network_performance_config": network_performance_config,
        "shielded_instance_integrity_policy": shielded_instance_integrity_policy,
        "private_ipv6_google_access": private_ipv6_google_access,
        "source_machine_image_encryption_key": source_machine_image_encryption_key,
        "confidential_instance_config": confidential_instance_config,
        "guest_accelerators": guest_accelerators,
        "description": description,
        "metadata": metadata,
        "resource_policies": resource_policies,
        "key_revocation_action_type": key_revocation_action_type,
        "hostname": hostname,
        "reservation_affinity": reservation_affinity,
        "tags": tags,
        "scheduling": scheduling,
        "advanced_machine_features": advanced_machine_features,
        "params": params,
        "machine_type": machine_type,
        "source_machine_image": source_machine_image,
        "shielded_instance_config": shielded_instance_config,
        "labels": labels,
        "deletion_protection": deletion_protection,
        "network_interfaces": network_interfaces,
        "can_ip_forward": can_ip_forward,
        "display_device": display_device,
        "zone": zone,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None
    if result["old_state"]:
        # copy.copy(old_get_ret['ret']) is needed to convert any objects of type NamespaceDict to dict
        # in old_get_ret['ret']. This is done so that comparisons with the resource_body which has
        # objects of type dict works properly
        zone = hub.tool.gcp.resource_prop_utils.parse_link_to_zone(
            result["old_state"]["zone"]
        )
        result["old_state"]["zone"] = zone

        # If rerun_data is set at this point this means that there has been a completed create or update operation
        # so there is no need to run the code for create or update below again and we stop here
        if ctx.get("rerun_data"):
            result["new_state"] = result["old_state"]
            return result

        # TODO: if we fix isPending() to not update after creation we can make the fingerprint required upon update
        resource_body["fingerprint"] = fingerprint or result["old_state"].get(
            "fingerprint"
        )
        resource_body["label_fingerprint"] = label_fingerprint or result[
            "old_state"
        ].get("label_fingerprint")

        # A dictionary of additional operations to perform on the object identified by the key
        # the values are tuples with the first element - the method to call, the second element - arguments, third - the property is required in the end result
        patch_operations_dict = {
            "network_interfaces": (
                hub.tool.gcp.compute.instance.update_network_interfaces,
                (ctx, result["old_state"], network_interfaces),
                True,
            ),
            "shielded_instance_config": (
                hub.tool.gcp.compute.instance.update_shielded_instance_config,
                (ctx, shielded_instance_config, None, None, None, resource_id),
                False,
            ),
            "shielded_instance_integrity_policy": (
                hub.tool.gcp.compute.instance.update_shielded_instance_integrity_policy,
                (
                    ctx,
                    shielded_instance_integrity_policy,
                    None,
                    None,
                    None,
                    resource_id,
                ),
                False,
            ),
        }

        old_properties_dict = {
            k: result["old_state"].get(k) for k in patch_operations_dict.keys()
        }
        changed_properties_dict = {
            k: hub.tool.gcp.utils.compare_states(
                {k: old_properties_dict.get(k, [])},
                {k: resource_body.get(k, [])},
                "compute.instance",
            )
            for k, v in patch_operations_dict.items()
        }

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            {"resource_id": resource_id, **resource_body, **old_properties_dict},
            "compute.instance",
        )

        if changes:
            changed_non_updatable_properties = (
                hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                    "compute.instance", changes
                )
            )
            if changed_non_updatable_properties:
                result["result"] = False
                result["comment"].append(
                    hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                        "gcp.compute.instance",
                        name,
                        changed_non_updatable_properties,
                    )
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
                return result

        if not changes and not any(changed_properties_dict.values()):
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_exists_comment(
                    "gcp.compute.instance", name
                )
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    "gcp.compute.instance", name
                )
            )
            result["new_state"] = resource_body
            result["new_state"]["resource_id"] = resource_id
            return result

        for k, op in patch_operations_dict.items():
            if changed_properties_dict[k]:
                ret = await op[0](*op[1])

                if not ret["result"]:
                    result["comment"] += ret["comment"]
                    result["result"] = False
                    return result

        if changes:
            if any(changed_properties_dict.values()):
                get_ret = await hub.exec.gcp.compute.instance.get(
                    ctx, resource_id=resource_id
                )
                if not get_ret["result"] and not get_ret["ret"]:
                    result["result"] = False
                    result["comment"] += get_ret["comment"]
                    return result

                for k in patch_operations_dict.keys():
                    resource_body[k] = get_ret["ret"].get(k)

                resource_body["fingerprint"] = get_ret["ret"].get("fingerprint")
                resource_body["label_fingerprint"] = get_ret["ret"].get(
                    "label_fingerprint"
                )

            # put the old value of required properties in the resource body
            for k, v in patch_operations_dict.items():
                if v[2] and not resource_body.get(k):
                    resource_body[k] = old_properties_dict[k]

            update_ret = await hub.exec.gcp_api.client.compute.instance.update(
                hub,
                ctx,
                name=name,
                resource_id=resource_id,
                minimal_action=minimal_action,
                most_disruptive_allowed_action=most_disruptive_allowed_action,
                body=resource_body,
            )

            if not update_ret["result"]:
                result["result"] = False
                result["comment"] += update_ret["comment"]
                return result
            if update_ret["ret"] is not None:
                if "compute#operation" in update_ret["ret"].get("kind", ""):
                    operation = update_ret["ret"]

        result["comment"].append(
            hub.tool.gcp.comment_utils.update_comment("gcp.compute.instance", name)
        )
    else:
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.instance", name
                )
            )
            result["new_state"] = resource_body
            result["new_state"]["resource_id"] = resource_id
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.compute.instance.insert(
            ctx,
            name=name,
            project=project,
            zone=zone,
            source_instance_template=source_instance_template,
            request_id=request_id,
            body=resource_body,
        )
        if not create_ret["result"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        "gcp.compute.instance", name
                    )
                )
                return result
            else:
                result["comment"] += create_ret["comment"]
        else:
            result["comment"].append(
                hub.tool.gcp.comment_utils.create_comment("gcp.compute.instance", name)
            )
            resource_id = create_ret["ret"].get("resource_id")
        if create_ret["ret"] is not None:
            if "compute#operation" in create_ret["ret"].get("kind", ""):
                operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.zone_operation"
        )
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, operation_id, "compute.instance"
        )

        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    # Try getting the resource again
    get_ret = await hub.exec.gcp.compute.instance.get(ctx, resource_id=resource_id)

    if not get_ret["result"] and not get_ret["ret"]:
        result["result"] = False
        result["comment"] += get_ret["comment"]
        return result

    result["new_state"] = get_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str = None,
    zone: str = None,
    project: str = None,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified Instance resource. For more information, see Deleting an instance.

    Args:
        name(str):
            The name of the resource

        resource_id(str, Optional):
            The resource_id of the resource

        zone:(str, Optional):
            The name of the zone for this request.
        project:(str, Optional):
            Project ID for this request.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed.
            For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported (00000000-0000-0000-0000-000000000000).

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        example_resource_name:
          gcp.compute.instance.absent

    """
    result = {
        "comment": [],
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    if not resource_id and not hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    ):
        resource_id = (ctx.get("old_state") or {}).get(
            "resource_id"
        ) or hub.tool.gcp.resource_prop_utils.construct_resource_id(
            "compute.instance", {**locals(), "instance": name}
        )

    if not resource_id:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                "gcp.compute.instance", name
            )
        )
        return result

    if not ctx.get("rerun_data"):
        get_ret = await hub.exec.gcp.compute.instance.get(ctx, resource_id=resource_id)

        if not get_ret["result"]:
            result["result"] = False
            result["comment"] += get_ret["comment"]
            return result

        if not get_ret["ret"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_absent_comment(
                    "gcp.compute.instance", name
                )
            )
            return result

        result["old_state"] = get_ret["ret"]

    if ctx["test"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                "gcp.compute.instance", name
            )
        )
        return result

    if not ctx.get("rerun_data"):
        # First iteration; invoke instance's delete()
        delete_ret = await hub.exec.gcp_api.client.compute.instance.delete(
            ctx, resource_id=resource_id, request_id=request_id
        )
        if "compute#operation" in delete_ret.get("ret", {}).get("kind"):
            result["result"] = False
            result["comment"] += delete_ret["comment"]
            result[
                "rerun_data"
            ] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                delete_ret["ret"].get("selfLink"), "compute.zone_operation"
            )
            return result
    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.instance"
        )
        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment("gcp.compute.instance", name)
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves the list of instances contained within the specified zone.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.instance

    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.instance.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe instances {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.instance.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



import dataclasses
from dataclasses import dataclass
from typing import Any
from typing import Dict

from idem_gcp.tool.gcp.generate.exec_param import ExecParam


@dataclass
class ExecutionContext:
    resource_type: str  # e.g. compute.instance
    method_name: str
    method_params: Dict[str, Any]
    exec_params: Dict[ExecParam, Any] = dataclasses.field(default_factory=dict)
    method: Any = None  # the resolved method to be executed
    response: Dict[str, Any] = None

    def __post_init__(self):
        self.fill_default_exec_params()

    def fill_default_exec_params(self):
        if self.exec_params is None:
            self.exec_params = dict()
        filled_exec_params = {**self.exec_params}
        for exec_param in ExecParam:
            if exec_param not in self.exec_params:
                filled_exec_params[exec_param] = exec_param.default_value
        self.exec_params = filled_exec_params

    def get_exec_param_value(self, param: ExecParam):
        return self.exec_params.get(param)

    def get_method_param_value(self, param: str):
        return self.method_params.get(param)



"""State module for managing Node Templates."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    request_id: str = None,
    project: str = None,
    region: str = None,
    description: str = None,
    disks: List[
        make_dataclass(
            "LocalDisk",
            [
                ("disk_count", int, field(default=0)),
                ("disk_type", str, field(default=None)),
                ("disk_size_gb", int, field(default=0)),
            ],
        )
    ] = None,
    cpu_overcommit_type: str = None,
    server_binding: make_dataclass(
        "ServerBinding",
        [
            ("type", str, field(default=None)),
        ],
    ) = None,
    node_affinity_labels: Dict = None,
    node_type_flexibility: make_dataclass(
        "NodeTemplateNodeTypeFlexibility",
        [
            ("memory", str, field(default=None)),
            ("local_ssd", str, field(default=None)),
            ("cpus", str, field(default=None)),
        ],
    ) = None,
    accelerators: List[
        make_dataclass(
            "AcceleratorConfig",
            [
                ("acceleratorType", str, field(default=None)),
                ("acceleratorCount", int, field(default=0)),
            ],
        )
    ] = None,
    node_type: str = None,
) -> Dict[str, Any]:
    r"""Creates a network in the specified project using the data included in the request.

    Args:
        name(str):
            An Idem name of the resource.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            The name of the region for this request.

        description(str, Optional):
            An optional description of this resource. Provide this field when you create the resource. Defaults to None.

        disks(List[LocalDisk], Optional):
            Local disk configurations.

        cpu_overcommit_type(str, Optional):
            CPU overcommit.

        server_binding(ServerBinding, Optional):
            Sets the binding properties for the physical server. Valid values include: - *[Default]* RESTART_NODE_ON_ANY_SERVER: Restarts VMs on any available physical server - RESTART_NODE_ON_MINIMAL_SERVER: Restarts VMs on the same physical server whenever possible See Sole-tenant node options for more information.

        node_affinity_labels({}, Optional):
            Labels to use for node affinity, which will be used in instance scheduling.

        node_type_flexibility(NodeTemplateNodeTypeFlexibility, Optional):
            The flexible properties of the desired node type. Node groups that use this node template will create nodes of a type that matches these properties. This field is mutually exclusive with the node_type property; you can only define one or the other, but not both.

        accelerators(List[AcceleratorConfig], Optional):
            A specification of the type and number of accelerator cards attached to the instance.

        node_type(str, Optional):
            The node type to use for nodes group that are created from this template.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = get_project_from_account(ctx, project)

    result["comment"].append(
        "No-op: There is no create/update function for gcp.compute.node_template"
    )

    return result


async def absent(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified network.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            delete-my-node-template:
              gcp.compute.node_template.absent:
                - resource_id: projects/project-name/regions/europe-west2/node_templates/node_template-123
    """
    result = {
        "comment": [],
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if ctx.test:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                "gcp.compute.node_template", name
            )
        )
        return result

    if not ctx.get("rerun_data"):
        # First iteration; invoke instance's delete()
        delete_ret = await hub.exec.gcp_api.client.compute.node_template.delete(
            ctx, resource_id=resource_id
        )
        if delete_ret["ret"]:
            if "compute#operation" in delete_ret["ret"].get("kind"):
                result["result"] = False
                result["comment"] += delete_ret["comment"]
                result[
                    "rerun_data"
                ] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                    delete_ret["ret"].get("selfLink"), "compute.region_operation"
                )
                return result

    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.node_template"
        )
        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    if not resource_id:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                "gcp.compute.node_template", name
            )
        )

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment("gcp.compute.node_template", name)
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves an aggregated list of node templates.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.node_template
    """
    result = {}

    # TODO: Pagination
    describe_ret = await hub.exec.gcp_api.client.compute.node_template.aggregatedList(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.node_template {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"]["items"]:
        resource_id = resource.get("resource_id")

        result[resource_id] = {
            "gcp.compute.node_template.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Exec module for managing Subnetworks."""
__func_alias__ = {"list_": "list"}

from idem_gcp.tool.gcp.utils import get_project_from_account


async def list_(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    filter: str = None,
    order_by: str = None,
):
    r"""Retrieves a list of subnetworks available to the specified project.

    Args:
        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            Name of the region scoping this request.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name.

    Examples:
        random-name:
          exec.run:
          - path: gcp.compute.subnetwork.list
          - kwargs:
              project: project-name
              region: region-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = get_project_from_account(ctx, project)

    if region:
        list_result = await hub.exec.gcp_api.client.compute.subnetwork.list(
            ctx,
            project=project,
            region=region,
            filter=filter,
            orderBy=order_by,
        )
    else:
        list_result = await hub.exec.gcp_api.client.compute.subnetwork.aggregatedList(
            ctx,
            project=project,
            filter=filter,
            orderBy=order_by,
        )

    result["comment"] += list_result["comment"]
    if not list_result["result"]:
        result["result"] = False
        return result

    items_list = None
    if list_result and list_result.get("ret"):
        items_list = list_result["ret"].get("items")

    result["ret"] = items_list
    return result


async def get(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    name: str = None,
    resource_id: str = None,
):
    r"""Returns the specified subnetwork. Gets a list of available subnetworks list() request.

    Args:
        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            Name of the region scoping this request. Defaults to None.

        name(str, Optional):
            Name of the Subnetwork resource to return. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Examples:
        random-name:
          exec.run:
          - path: gcp.compute.subnetwork.get
          - kwargs:
              project: project-name
              region: region-name
              name: subnetwork-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.subnetwork.get(
            ctx,
            resource_id=resource_id,
        )
    elif project and region and name:
        ret = await hub.exec.gcp_api.client.compute.subnetwork.get(
            ctx, project=project, region=region, subnetwork=name, name=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.subnetwork {name} either resource_id or project, region and name"
            f" should be specified."
        ]
        return result

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result



"""Google API Python client library exec module.

Copyright (c) 2021-2022 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0

This file implements the entirety of the Plugin Oriented Programming (POP)
exec Sub for Google discovery-based Python client library APIs.

The basic operating model is building Subs as needed to match the nature of a
call to an exec. For example, a call like:

    hub.exec.gcp.compute.instance.get(ctx, ...)

will work, even though no exec subdirectories exist to match that call path.
Instead, the API is dynamically discovered and built to match and tie
them to the appropriate GCP Python SDK wrappers within the tool Sub of this
project.
"""
import copy
from inspect import signature
from typing import Dict
from typing import List

from pop.loader import LoadedMod

from idem_gcp.helpers.exc import NotFoundError

_GCP_SUB_CACHE = {}
_GCP_RESOURCE_TYPES_CACHE = {}


def resolve_sub(hub, ref: str, gcp_api):
    """
    Produces a callable from a string (dotted notation) call to an API.
    :param hub: The Hub into which the resolved callable will get placed.
    :param ref: The call path reference. For example:
        "hub.exec.gcp.compute.instance.get".
    """

    def gen_sub(ref: str, gcp_api):
        callpath = ref.split(".")
        resource_type = ".".join(callpath[1:-1])
        native_resource_type = hub.tool.gcp.resolver.find_native_resource_type(
            resource_type
        )
        method_name = callpath[-1]
        callpath = callpath[0:1] + native_resource_type.split(".") + [callpath[-1]]
        callpath[0:-1] = [hub.tool.gcp.case.camel(v) for v in callpath[0:-1]]

        gcp_api_sub = gcp_api
        for c in callpath[1:]:
            gcp_api_sub = getattr(gcp_api_sub, c)
        cm = ContractMod(
            hub, gcp_api_sub, resource_type, native_resource_type, method_name
        )
        _GCP_SUB_CACHE[ref] = cm
        return cm

    return _GCP_SUB_CACHE.get(ref, gen_sub(ref, gcp_api))


class ContractMod(LoadedMod):
    """
    A Sub (class) to represent actually callable methods within Azure
    ...ManagementClient API sets.

    For example, "list" within the virtual_machines API set in
    ComputeManagementClient.
    """

    def __init__(
        self, hub, gcp_method, resource_type, native_resource_type, method_name
    ):
        """
        Initialize the instance.
        :param hub: The redistributed pop central hub to which to attach
        the Subs created by this class.
        :param gcp_method: The _GCPApi representing the method (from the
        tool Sub).
        """
        super().__init__(name=gcp_method.name)
        self._hub = hub
        self._gcp_method = gcp_method
        self._resource_type = resource_type
        self._native_resource_type = native_resource_type
        self._method_name = method_name

    @property
    def signature(self):
        """
        Returns the signature of the __call__ method.

        When calling functions that contain "ctx" within the arguments to the
        function, POP injects the acct profile into "ctx" so it needs the
        call signature in order to do that.
        """
        return signature(self.__call__)

    def _missing(self, item: str):
        """Return a value for the given key (item)."""
        return self

    async def __call__(self, ctx, *args, **kwargs):
        """
        Closure on hub/target that calls the target function.
        :param ctx: A dict with the keys/values for the execution of the Idem run
        located in `hub.idem.RUNS[ctx['run_name']]`.
        :param args: Tuple of positional arguments.
        :param kwargs: dict of named *keyward/value) arguments.
        :return: Results of call in standard exec packaging:
            result: True if the API succeeded, False otherwise.
            ret: Any data returned by the API.
            comment: Any relevant info generated by the API (e.g.,
            "code 200", "code 404", an exception message, etc.).
        """
        result = {"result": True, "ret": None, "comment": []}

        # TODO: Set generation and metageneration

        present_params = copy.deepcopy(kwargs)
        name = present_params.pop("name", None)
        resource_id = present_params.pop("resource_id", None)

        if resource_id:
            present_params.update(
                self._hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
                    self._resource_type, resource_id
                )
            )

        if present_params.get("user_project") == "None":
            present_params["user_project"] = None

        exclude_keys_from_transformation = None
        if present_params.get("body", False):
            exclude_keys_from_transformation = self._hub.tool.gcp.resource_prop_utils.get_exclude_keys_from_transformation(
                present_params["body"], self._resource_type
            )

        raw_params = (
            self._hub.tool.gcp.conversion_utils.convert_present_resource_to_raw(
                present_params, self._resource_type, exclude_keys_from_transformation
            )
        )

        # TODO: Fix projection required or not
        if self._method_name in {"create"} and not self._resource_type.startswith(
            "cloudkms"
        ):
            raw_params["projection"] = "full"

        try:
            response = await self._gcp_method(ctx, *args, **raw_params)
            # Is it possible to get None for response. It looks like on error exceptions is thrown. If we are here
            # then the call was successful.
            if response is None:
                result["result"] = False
                return result

            if "#operation" in response.get("kind", ""):
                result["ret"] = response
                return result

            items = response.get("items")
            if isinstance(items, dict):
                response[
                    "items"
                ] = self._hub.tool.gcp.resolver.convert_dict_items_to_list(
                    items, self._native_resource_type
                )

            resource_type_camel = self._hub.tool.gcp.case.camel(
                self._native_resource_type.split(".")[-1]
            )
            if resource_type_camel in response:
                response["items"] = response[resource_type_camel]

            # TODO: Handle regional and global operations

            if response.get("nextPageToken"):
                all_items = response.get("items", [])
                raw_params["pageToken"] = response["nextPageToken"]

                while response.get("nextPageToken"):
                    response = await self._gcp_method(ctx, *args, **raw_params)

                    items = response.get("items")
                    if isinstance(items, dict):
                        response[
                            "items"
                        ] = self._hub.tool.gcp.resolver.convert_dict_items_to_list(
                            items, self._native_resource_type
                        )

                    resource_type_camel = self._hub.tool.gcp.case.camel(
                        self._native_resource_type.split(".")[-1]
                    )
                    if resource_type_camel in response:
                        response["items"] = response[resource_type_camel]

                    all_items += response.get("items", [])
                    raw_params["pageToken"] = response.get("nextPageToken")

                response["items"] = all_items

            result["ret"] = self._hub.tool.gcp.resolver.process_response(
                response, self._resource_type, self._method_name
            )
        except Exception as e:
            if self._method_name == "get" and isinstance(e, NotFoundError):
                result["comment"].append(
                    self._hub.tool.gcp.comment_utils.get_empty_comment(
                        self._resource_type, name if name else resource_id
                    )
                )
            else:
                result["result"] = False
            result["comment"].append(str(e))

        return result


def process_response(hub, response: Dict, resource_type: str, method_name: str) -> Dict:
    # TODO: Get does not return all the necessary properties which should be the present properties so
    #  decide what to do

    present_props_names = hub.tool.gcp.resource_prop_utils.get_present_properties(
        resource_type
    )
    singular_result_expected = hub.tool.gcp.resolver.has_singular_result(
        response, resource_type, method_name
    )
    if singular_result_expected:
        resource_id = hub.tool.gcp.resource_prop_utils.extract_resource_id(
            response, resource_type
        )
        if resource_id:
            present_item = hub.tool.gcp.resolver.filter_raw_properties_to_present(
                response, present_props_names, resource_type, resource_id
            )
            return present_item

        return response
    else:
        # Multiple entries in call result, e.g. after list() call
        ret_items = []
        for service_resource in response.get("items") or {}:
            resource_id = hub.tool.gcp.resource_prop_utils.extract_resource_id(
                service_resource, resource_type
            )

            present_item = None
            if resource_id:
                present_item = hub.tool.gcp.resolver.filter_raw_properties_to_present(
                    service_resource, present_props_names, resource_type, resource_id
                )

            ret_items.append(present_item if present_item else service_resource)

        return {"items": ret_items}

    return {}


def filter_raw_properties_to_present(
    hub,
    raw_props: Dict,
    present_properties_names: List,
    resource_type: str,
    resource_id: str,
) -> Dict:
    filtered = {
        key: value
        for key, value in raw_props.items()
        if key in present_properties_names
    }
    filtered["resource_id"] = resource_id
    filtered = hub.tool.gcp.resource_prop_utils.format_path_params(
        filtered, resource_type
    )
    return hub.tool.gcp.conversion_utils.convert_raw_resource_to_present(filtered)


def convert_dict_items_to_list(hub, dict_items: Dict, resource_type: str) -> list:
    list_items = []
    simple_resource_type = resource_type.split(".")[-1]
    simple_resource_type_camel = hub.tool.gcp.case.camel(simple_resource_type)

    # TODO: How to check that for every resource the nested container in grouped list results is always
    #  named after the resource itself?
    for group_name in dict_items.values():
        if (
            simple_resource_type not in group_name
            and simple_resource_type_camel not in group_name
        ):
            continue
        for service_resource in group_name.get(simple_resource_type_camel):
            list_items.append(service_resource)

    return list_items


def has_singular_result(
    hub, response: Dict, resource_type: str, method_name: str
) -> bool:
    # TODO: Figure out a better mechanism to determine whether a given method is expected to return a single entity
    if isinstance(response.get("items"), list):
        return False
    if method_name in {"list", "aggregatedList", "describe"}:
        return False
    return True


def find_native_resource_type(hub, resource_type: str) -> str:
    def fetch_res_type_mapping(res_type: str) -> str:
        hub_ref = hub.metadata.gcp
        for part in res_type.split("."):
            hub_ref = hub_ref[part]
        try:
            return hub_ref["NATIVE_RESOURCE_TYPE"]
        except AttributeError:
            return res_type

    return _GCP_RESOURCE_TYPES_CACHE.get(
        resource_type, fetch_res_type_mapping(resource_type)
    )



"""State module for managing Buckets."""
import copy
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import get_project_from_account

__contracts__ = ["resource"]


# TODO:
# TREQ = {
#     "present": {
#
#     }
# }


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    predefined_acl: str = None,
    predefined_default_object_acl: str = None,
    project: str = None,
    # TODO: Remove user_project because it cannot be changed as it is set in the credentials
    user_project: str = None,
    acl: List = None,
    billing: Dict = None,
    cors: List = None,
    custom_placement_config: Dict = None,
    default_event_based_hold: bool = None,
    default_object_acl: List = None,
    encryption: Dict = None,
    iam_configuration: Dict = None,
    labels: Dict = None,
    lifecycle: Dict = None,
    location: str = None,
    logging: Dict = None,
    rpo: str = None,
    storage_class: str = None,
    versioning: Dict = None,
    website: Dict = None,
):
    """Ensure a bucket exists pursuant to the requested state.

    :param hub: The redistributed pop central hub.
    :param ctx: A dict with the keys/values for the execution of the Idem run
    located in `hub.idem.RUNS[ctx['run_name']]`.
    :param name: The name of the state (e.g., "assure bucket present").
    :param project: Name of the project in which to insert/update the bucket.
    :param userProject: Project id of the user to bill.
    :param body: A `Bucket`_ dictionary passed as the body to the GCP APIs
    as they require.

    Example SLS file usage:
    .. code-block::
        "Assure Bucket Present":
            gcp.storage.buckets.present:
                - project: MyProject
                - userProject: None
                - body:
                    name: MyBucket
                    cors:
                    - maxAgeSeconds: 42
                      method:
                        - "GET"
                        - "OPTIONS"
                        - "POST"
                        - "PATCH"
                        - "DELETE"
                      origin": [ '*' ]
                      responseHeader:
                        - "Cache-Control"
                        - "Content-Language"
                        - "Content-Length"
                        - "Content-Type"
                        - "Expires"
                        - "Last-Modified"
                        - "Pragma"

    .. _Bucket: https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.buckets.html#insert
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = get_project_from_account(ctx, project)

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    old = None

    if resource_id:
        old_get_ret = await hub.exec.gcp.storage.buckets.get(
            ctx,
            name=name,
        )

        if not old_get_ret["result"] or not old_get_ret["ret"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            # TODO: Handle the case when resource does not exist
            return result

        # copy.copy(old_get_ret['ret']) is needed to convert any objects of type NamespaceDict to dict
        # in old_get_ret['ret']. This is done so that comparisons with the plan_state which has
        # objects of type dict works properly
        old = copy.deepcopy(copy.copy(old_get_ret["ret"]))
        result["old_state"] = old

    # TODO: Check if body contains the same parameters as old
    # to be autogenerated by pop-create based on insert/update props in properties.yaml
    resource_body = {
        "acl": acl,
        "billing": billing,
        "cors": cors,
        "custom_placement_config": custom_placement_config,
        "default_event_based_hold": default_event_based_hold,
        "default_object_acl": default_object_acl,
        "encryption": encryption,
        "iam_configuration": iam_configuration,
        "labels": labels,
        "lifecycle": lifecycle,
        "location": location,
        "logging": logging,
        "name": name,
        "rpo": rpo,
        "storage_class": storage_class,
        "versioning": versioning,
        "website": website,
    }

    # TODO: How to handle query params which are not returned in the response body of get
    plan_state = {
        # "project": project,
        # "predefined_acl": predefined_acl,
        # "predefined_default_object_acl": predefined_default_object_acl,
        "resource_id": resource_id,
        **resource_body,
    }

    plan_state = {k: v for (k, v) in plan_state.items() if v is not None}

    if old:
        changes = hub.tool.gcp.utils.compare_states(old, plan_state)

        if not changes:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_exists_comment(
                    "gcp.storage.bucket", name
                )
            )
            result["new_state"] = plan_state

            return result

        non_updatable_properties = (
            hub.tool.gcp.resource_prop_utils.get_non_updatable_properties(
                "storage.buckets"
            )
        )

        for key, value in plan_state.items():
            if value == old.get(key):
                non_updatable_properties.discard(key)

        if non_updatable_properties:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                    "gcp.storage.bucket", name, non_updatable_properties
                )
            )
            result["new_state"] = copy.deepcopy(old)
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update("gcp.storage.bucket", name)
            )
            result["new_state"] = plan_state
            return result

        # Perform update
        update_ret = await hub.exec.gcp_api.client.storage.buckets.update(
            hub,
            ctx,
            name=name,
            resource_id=resource_id,
            body=resource_body,
            userProject=user_project,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] += update_ret["comment"]
            return result
        result["comment"].append(
            hub.tool.gcp.comment_utils.update_comment("gcp.storage.bucket", name)
        )
        result["new_state"] = update_ret["ret"]

    else:
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.storage.bucket", name
                )
            )
            result["new_state"] = plan_state
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.storage.buckets.insert(
            ctx,
            name=name,
            project=project,
            body=resource_body,
            userProject=user_project,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
            return result
        result["comment"].append(
            hub.tool.gcp.comment_utils.create_comment("gcp.storage.bucket", name)
        )
        result["old_state"] = {}
        result["new_state"] = create_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    project: str = None,
    user_project: str = None,
):
    """Ensure a bucket does not exist.

    :param hub: The redistributed pop central hub.
    :param ctx: A dict with the keys/values for the execution of the Idem run
    located in `hub.idem.RUNS[ctx['run_name']]`.
    :param name: The name of the state (e.g., "assure subnet absent").
    :param project: The project in which the bucket should not exist.

    Example usage:
        .. code-block: yaml
            gcp.storage.buckets.absent:
                - project: {{ project }}
                - bucket: {{ bucket_name }}
                - user_project: {{ billto_project_name }}
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if not resource_id:
        result["comment"].append(
            hub.tool.aws.comment_utils.already_absent_comment(
                "gcp.storage.buckets", name
            )
        )
        return result

    get_ret = await hub.exec.gcp.storage.buckets.get(ctx, name=name)
    if not get_ret["result"] or not get_ret["ret"]:
        result["result"] = get_ret["result"]
        result["comment"] += get_ret["comment"]
        # TODO: Should we get the old state from the context?
        return result

    if ctx["test"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment("gcp.storage.bucket", name)
        )
        result["old_state"] = get_ret["ret"]
        return result
    else:
        # Delete the existing bucket (async call).
        del_ret = await hub.exec.gcp.storage.buckets.delete(
            ctx, name=name, resource_id=resource_id, userProject=user_project
        )
        if not del_ret["result"]:
            result["result"] = False
            result["comment"] += del_ret["comment"]
            return result

        result["old_state"] = get_ret["ret"]
        result["comment"].append(
            hub.tool.gcp.comment_utils.delete_comment("gcp.storage.bucket", name)
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of buckets.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.storage.buckets
    """
    result = {}

    # TODO: Pagination
    describe_ret = await hub.exec.gcp.storage.buckets.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe buckets {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]["items"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.storage.buckets.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



def create_comment(hub, resource_type: str, name: str) -> str:
    return f"Created {resource_type} '{name}'"


def would_create_comment(hub, resource_type: str, name: str) -> str:
    return f"Would create {resource_type} '{name}'"


def update_comment(hub, resource_type: str, name: str) -> str:
    return f"Updated {resource_type} '{name}'"


def would_update_comment(hub, resource_type: str, name: str) -> str:
    return f"Would update {resource_type} '{name}'"


def no_resource_delete_comment(hub, resource_type: str) -> str:
    return f"Delete operation for resource type {resource_type} is not supported."


def no_resource_create_update_comment(hub, resource_type: str) -> str:
    return f"Create and update operations for resource type {resource_type} are not supported."


def delete_comment(hub, resource_type: str, name: str) -> str:
    return f"Deleted {resource_type} '{name}'"


def would_delete_comment(hub, resource_type: str, name: str) -> str:
    return f"Would delete {resource_type} '{name}'"


def already_absent_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' already absent"


def already_exists_comment(hub, resource_type: str, name: str) -> str:
    return f"{resource_type} '{name}' already exists."


def update_tags_comment(hub, tags_to_remove, tags_to_add) -> str:
    return f"Update tags: Add keys {tags_to_add.keys()} Remove keys {tags_to_remove.keys()}"


def would_update_tags_comment(hub, tags_to_remove, tags_to_add) -> str:
    return f"Would update tags: Add keys {tags_to_add.keys()} Remove keys {tags_to_remove.keys()}"


def get_empty_comment(hub, resource_type: str, name: str) -> str:
    return f"Get {resource_type} '{name}' result is empty"


def list_empty_comment(hub, resource_type: str, name: str) -> str:
    return f"List {resource_type} '{name}' result is empty"


def find_more_than_one(hub, resource_type: str, resource_id: str) -> str:
    return (
        f"More than one {resource_type} resource was found. Use resource {resource_id}"
    )


def non_updatable_properties_comment(
    hub, resource_type: str, name: str, non_updatable_properties: set
) -> str:
    sorted_props = list(non_updatable_properties)
    sorted_props.sort()
    return f"Forbidden modification of non-updatable properties: {str(sorted_props)} for {resource_type} '{name}'"


def no_resource_update_comment(hub, resource_type: str, resource_id: str) -> str:
    return f"Update operation for resource type {resource_type} is not supported for {resource_id}."


def ill_formed_resource_id_comment(
    hub, resource_type: str, resource_id: str, reason: str
) -> str:
    return f"Ill-formed {resource_type} resource ID '{resource_id}': {reason}"



import copy
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Set

import yaml
from deepdiff import DeepDiff
from importlib_resources import files


def __init__(hub):
    file_text = (
        files("idem_gcp.resources")
        .joinpath("properties.yaml")
        .read_text(encoding="utf-8")
    )

    # TODO: Maybe convert the properties names in hub.tool.gcp.RESOURCE_PROPS
    #  to snake case here so that we do not have to do conversion
    #  very often anywhere these snake case properties names are needed.
    #  Check if hub.tool.gcp.RESOURCE_PROPS are always convert to snake case

    hub.tool.gcp.RESOURCE_PROPS = yaml.safe_load(file_text)


def get_create_properties(hub, resource_type: str, convert_to_present=True) -> Set:
    resource_methods_properties = hub.tool.gcp.RESOURCE_PROPS[resource_type]
    raw_props = set(
        resource_methods_properties.get("insert", {})
        or resource_methods_properties.get("create", {})
    )
    if raw_props and convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_update_properties(hub, resource_type: str, convert_to_present=True) -> Set:
    raw_props = set(hub.tool.gcp.RESOURCE_PROPS[resource_type].get("update", {}))
    if convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_fields_of_enum_type(hub, resource_type: str, convert_to_present=True) -> Set:
    raw_props = set(
        hub.tool.gcp.RESOURCE_PROPS[resource_type].get("fields_enum_type", {})
    )
    if convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_get_properties(hub, resource_type: str, convert_to_present=True) -> Set:
    raw_props = set(hub.tool.gcp.RESOURCE_PROPS[resource_type].get("get", {}))
    if convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_delete_properties(hub, resource_type: str, convert_to_present=True) -> Set:
    raw_props = set(hub.tool.gcp.RESOURCE_PROPS[resource_type].get("delete", {}))
    if convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_non_updatable_properties(hub, resource_type: str) -> Set:
    raw_nested_non_updatable_properties = hub.tool.gcp.RESOURCE_PROPS[
        resource_type
    ].get("nested_non_updatable_properties", [])

    non_updatable_properties = set(
        hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_nested_non_updatable_properties
        )
    )

    non_updatable_properties.update(
        hub.tool.gcp.resource_prop_utils.get_create_properties(resource_type)
        - hub.tool.gcp.resource_prop_utils.get_update_properties(resource_type)
    )
    return non_updatable_properties


def get_changed_non_updatable_properties(
    hub, resource_type: str, changes: DeepDiff
) -> Set:
    non_updatable_properties = (
        hub.tool.gcp.resource_prop_utils.get_non_updatable_properties(resource_type)
    )

    changed_non_updatable_properties = set()
    for changed_property_full_path in changes.affected_paths:
        # This formats changed_property_full_path from e.g. "root['network_interfaces'][0]['alias_ip_ranges']"
        # to "network_interfaces[].alias_ip_ranges"
        changed_property_path_elements = re.findall(
            r"(?<=\[).*?(?=\])", changed_property_full_path
        )
        formatted_path_elements = []
        for i in range(len(changed_property_path_elements)):
            if changed_property_path_elements[i].isdigit():
                formatted_path_elements.append("[]")
            else:
                if i > 0:
                    formatted_path_elements.append(".")
                formatted_path_elements.append(
                    changed_property_path_elements[i].strip("'")
                )
        formatted_property_path = "".join(formatted_path_elements)
        for non_updatable_property_path in non_updatable_properties:
            if non_updatable_property_path in formatted_property_path:
                changed_non_updatable_properties.add(non_updatable_property_path)
                break

    return changed_non_updatable_properties


def get_present_properties(hub, resource_type: str) -> Set:
    return (
        hub.tool.gcp.resource_prop_utils.get_create_properties(resource_type, False)
        .union(
            hub.tool.gcp.resource_prop_utils.get_update_properties(resource_type, False)
        )
        .union(
            hub.tool.gcp.resource_prop_utils.get_fields_of_enum_type(
                resource_type, False
            )
        )
    )


def get_exclude_paths(hub, resource_type: str) -> Set:
    raw_props_names = hub.tool.gcp.RESOURCE_PROPS[resource_type].get(
        "exclude_paths", []
    )
    return set(
        hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(raw_props_names)
    )


def get_exclude_properties_from_transformation(
    hub, resource_type: str, convert_to_present=True
) -> Set:
    resource_methods_properties = hub.tool.gcp.RESOURCE_PROPS[resource_type]
    raw_props = set(
        resource_methods_properties.get("exclude_properties_from_transformation", {})
    )
    if raw_props and convert_to_present:
        return hub.tool.gcp.conversion_utils.convert_raw_properties_to_present(
            raw_props
        )
    return raw_props


def get_exclude_keys_from_transformation(
    hub, resource_body: Dict[str, Any], resource_type: str = None
) -> List[str]:
    # Get the properties whose keys should be excluded from transformation
    exclude_properties_from_transformation = (
        hub.tool.gcp.resource_prop_utils.get_exclude_properties_from_transformation(
            resource_type
        )
    )

    # Create a list, which has lists per each property and have the properties' keys as items
    list_of_properties_key_lists = [
        list(resource_body[item].keys()) if resource_body.get(item, False) else []
        for item in exclude_properties_from_transformation
    ]

    # Map the list_of_properties_key_lists to a list containing only the keys to be excluded
    exclude_keys_from_transformation = []
    for _list in list_of_properties_key_lists:
        exclude_keys_from_transformation += _list

    return exclude_keys_from_transformation


def extract_resource_id(hub, input_props: Dict, resource_type: str) -> Dict:
    if "selfLink" in input_props:
        return hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            input_props.get("selfLink"), resource_type
        )

    if "resource_id" in hub.tool.gcp.RESOURCE_PROPS[resource_type]:
        resource_id_fields = hub.tool.gcp.RESOURCE_PROPS[resource_type]["resource_id"]
        for resource_id_field in resource_id_fields:
            if resource_id_field in input_props:
                return input_props[resource_id_field]


def get_resource_path(hub, resource_type: str) -> str:
    resource_path = resource_type.split(".")
    hub_ref = hub.metadata.gcp
    for resource_path_segment in resource_path:
        hub_ref = hub_ref[resource_path_segment]

    return hub_ref["PATH"]


def get_resource_id_property_name(hub, resource_type: str, method_name: str) -> str:
    resource_id_props = hub.tool.gcp.RESOURCE_PROPS[resource_type]["resource_id"]
    method_properties = {}
    if method_name == "get":
        method_properties = hub.tool.gcp.resource_prop_utils.get_get_properties(
            resource_type
        )
    elif method_name == "insert" or method_name == "create":
        method_properties = hub.tool.gcp.resource_prop_utils.get_create_properties(
            resource_type
        )
    elif method_name == "update":
        method_properties = hub.tool.gcp.resource_prop_utils.get_update_properties(
            resource_type
        )
    elif method_name == "delete":
        method_properties = hub.tool.gcp.resource_prop_utils.get_delete_properties(
            resource_type
        )

    for resource_id_prop in resource_id_props:
        if resource_id_prop in method_properties:
            return resource_id_prop

    return ""


def get_elements_from_resource_id(hub, resource_type: str, resource_id: str) -> Dict:
    resource_path = hub.tool.gcp.resource_prop_utils.get_resource_path(resource_type)
    src = resource_path.split("/")
    act = resource_id.split("/")
    result = {}

    if len(act) < len(src):
        hub.log.warning(
            hub.tool.gcp.comment_utils.ill_formed_resource_id_comment(
                resource_type, resource_id, "insufficient number of segments."
            )
        )
        return {}

    idx = -1
    for s in reversed(src):
        val = act[idx]
        if s.startswith("{") and s.endswith("}"):
            key = s[1:-1]
            result[key] = val
        elif val != s:
            hub.log.warning(
                hub.tool.gcp.comment_utils.ill_formed_resource_id_comment(
                    resource_type, resource_id, "segment names do not match."
                )
            )
            return {}
        idx -= 1

    return result


def parse_link_to_resource_id(hub, link, resource_type: str):
    resource_path = hub.tool.gcp.resource_prop_utils.get_resource_path(resource_type)
    result = copy.deepcopy(resource_path)
    src = result.split("/")
    act = link.split("/")

    if len(act) < len(src):
        return None

    idx = -1
    for s in reversed(src):
        val = act[idx]
        if s.startswith("{") and s.endswith("}"):
            result = result.replace(s, val)
        elif val != s:
            hub.log.warning(
                f"Resource selfLink differs from the actual path for resource type: {resource_type}"
                f"\nselfLink: {link}, PATH: {resource_path}"
            )
            return None
        idx -= 1

    return result


def construct_resource_id(hub, resource_type: str, input_props: Dict[str, Any]) -> str:
    if input_props is not None and input_props.get("name") is not None:
        # handle case where "name" contains full resource id
        name = input_props.get("name")
        if (
            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                name, resource_type
            )
            is not None
        ):
            return name

    resource_path = hub.tool.gcp.resource_prop_utils.get_resource_path(resource_type)
    try:
        filtered_without_none_properties = {
            k: v for k, v in input_props.items() if v is not None
        }
        return resource_path.format(**filtered_without_none_properties)
    except:
        # a required input prop is missing or input_props is invalid format
        return None


def parse_link_to_zone(hub, link: str) -> str:
    return link.split("/")[-1]


def get_path_parameters(hub, resource_type):
    resource_path = hub.tool.gcp.resource_prop_utils.get_resource_path(resource_type)
    src = resource_path.split("/")
    result = set()

    for s in src:
        if s.startswith("{") and s.endswith("}"):
            result.add(s[1:-1])

    return result


def format_path_params(hub, resource, resource_type):
    resource_copy = copy.deepcopy(resource)
    path_params = hub.tool.gcp.resource_prop_utils.get_path_parameters(resource_type)
    for path_el in path_params:
        if resource_copy.get(path_el):
            resource_copy[path_el] = resource_copy[path_el].split("/")[-1]

    return resource_copy


# TODO: Remove this logic once we have all the nested properties defined for a method
def are_properties_allowed_for_update(hub, resource_type, request_body):
    can_update = True
    if resource_type == "compute.instance":
        for disk in request_body.get("disks" or {}):
            if disk.get("initialize_params"):
                can_update = False

    return can_update


def resource_type_matches(hub, resource_id: str, resource_type: str) -> bool:
    resource_path = hub.tool.gcp.resource_prop_utils.get_resource_path(resource_type)
    rpath = resource_path.split("/")
    rid = resource_id.split("/")

    if len(rid) < len(rpath):
        return False

    idx = -1
    for s in reversed(rpath):
        if not (s.startswith("{") and s.endswith("}")) and rid[idx] != s:
            return False
        idx -= 1

    return True



"""Metadata module for managing Object Access Controls."""
# this will be autogenerated by pop-create with the resource id based on parameterOrder value from JSON schema for GET
RESOURCE_IDS = ["bucket", "object", "entity"]



"""BCP Resource Manager (AzureRM) exec contracts.

Copyright (c) 2021 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""


async def call(hub, ctx):
    """Override the default call contract in order to prevent Exception usurping by Idem.

    :param hub: The redistributed pop central hub.
    :param ctx: A dict with the keys/values  for the execution of the Idem run
    located in `hub.idem.RUNS[ctx['run_name']]`.
    """
    return await ctx.func(*ctx.args, **ctx.kwargs)



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_network_interfaces(
    hub,
    ctx,
    current_instance: Dict[str, Any],
    new_network_interfaces: List[Dict[str, Any]],
) -> Dict[str, Any]:
    # Update the updatable network interface properties and the access configs
    result = {"result": False, "comment": []}

    existing_network_interfaces = copy.deepcopy(
        current_instance.get("network_interfaces", [])
    )
    for net_intf in new_network_interfaces:
        existing_net_intf = next(
            (
                existing_network_interfaces[idx]
                for idx in range(len(existing_network_interfaces))
                if existing_network_interfaces[idx].get("name") == net_intf.get("name")
            ),
            None,
        )
        if not existing_net_intf:
            break

        # According to documentation, only one access config per instance is supported.

        existing_access_configs = existing_net_intf.get("access_configs", [])
        desired_access_configs = net_intf.get("access_configs", [])

        existing_ac = (
            existing_access_configs[0] if len(existing_access_configs) > 0 else None
        )
        desired_ac = (
            desired_access_configs[0] if len(desired_access_configs) > 0 else None
        )

        if not existing_ac and not desired_ac:
            continue

        elif existing_ac and not desired_ac:
            op_ret = await hub.exec.gcp_api.client.compute.instance.deleteAccessConfig(
                ctx,
                resource_id=current_instance.get("resource_id"),
                access_config=existing_ac.get("name"),
                network_interface=net_intf.get("name"),
            )

            r = await hub.tool.gcp.operation_utils.await_operation_completion(
                ctx, op_ret, "compute.instance", "compute.zone_operation"
            )
            if not r["result"]:
                result["comment"] += r["comment"]
                return result

        elif desired_ac and not existing_ac:
            op_ret = await hub.exec.gcp_api.client.compute.instance.addAccessConfig(
                ctx,
                resource_id=current_instance.get("resource_id"),
                network_interface=net_intf.get("name"),
                body=desired_ac,
            )

            r = await hub.tool.gcp.operation_utils.await_operation_completion(
                ctx, op_ret, "compute.instance", "compute.zone_operation"
            )
            if not r["result"]:
                result["comment"] += r["comment"]
                return result

        # existing_ac and desired_ac
        elif next(
            (
                key
                for key in desired_ac.keys()
                if desired_ac.get(key) != existing_ac.get(key)
            ),
            None,
        ):
            op_ret = await hub.exec.gcp_api.client.compute.instance.updateAccessConfig(
                ctx,
                resource_id=current_instance.get("resource_id"),
                network_interface=net_intf.get("name"),
                body=desired_ac,
            )

            r = await hub.tool.gcp.operation_utils.await_operation_completion(
                ctx, op_ret, "compute.instance", "compute.zone_operation"
            )
            if not r["result"]:
                result["comment"] += r["comment"]
                return result

        # Network interface update
        # Cannot update access config through network interface update function - must be a separate call
        body = hub.tool.gcp.utils.create_dict_body_on_top_of_old(
            ctx, existing_net_intf, net_intf
        )

        # We want to exclude access configs from update nic operation because it is not supported and
        # it is handled separately.
        body.pop("access_configs", None)
        existing_net_intf.pop("access_configs", None)

        if body != existing_net_intf:
            op_ret = (
                await hub.exec.gcp_api.client.compute.instance.updateNetworkInterface(
                    ctx,
                    resource_id=current_instance.get("resource_id"),
                    network_interface=existing_net_intf.get("name"),
                    body=body,
                )
            )

            ret = await hub.tool.gcp.operation_utils.await_operation_completion(
                ctx, op_ret, "compute.instance", "compute.zone_operation"
            )
            if not ret["result"]:
                result["comment"] += ret["comment"]
                return result

    result["result"] = True
    return result


async def update_shielded_instance_config(
    hub,
    ctx,
    shielded_instance_config: Dict[str, bool],
    project: str = None,
    zone: str = None,
    instance: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    return await hub.exec.gcp.compute.instance.update_shielded_instance_config(
        ctx,
        shielded_instance_config["enable_secure_boot"],
        shielded_instance_config["enable_vtpm"],
        shielded_instance_config["enable_integrity_monitoring"],
        project,
        zone,
        instance,
        resource_id,
    )


async def update_shielded_instance_integrity_policy(
    hub,
    ctx,
    shielded_instance_integrity_policy: Dict[str, bool],
    project: str = None,
    zone: str = None,
    instance: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    return await hub.exec.gcp.compute.instance.set_shielded_instance_integrity_policy(
        ctx,
        shielded_instance_integrity_policy["update_auto_learn_policy"],
        project,
        zone,
        instance,
        resource_id,
    )



from typing import List


def are_lists_identical(hub, list1: List, list2: List) -> bool:
    if not list1 and not list2:
        return True
    if list1 is None or len(list1) == 0 or list2 is None or len(list2) == 0:
        return False

    for l in [list1, list2]:
        if not isinstance(l, List):
            raise TypeError(
                f"Expecting lists to compare. This is expected to be of type List: '{l}'"
            )

    diff = [i for i in list1 + list2 if i not in list1 or i not in list2]
    result = len(diff) == 0
    if not result:
        hub.log.debug(f"There are {len(diff)} differences.")
    return result
